"""Tests for tdnet-disclosure-mcp server."""

from datetime import datetime
from unittest.mock import Mock, patch

import pytest

import tdnet_disclosure_mcp.server as server_module
from tdnet_disclosure_mcp.models import (
    Disclosure,
    DisclosureCategory,
    DisclosureList,
)


class TestServerHelpers:
    """Test server helper functions."""

    @pytest.mark.asyncio
    async def test_get_client_creates_client(self) -> None:
        with (
            patch.object(server_module, "_client", None),
            patch("tdnet_disclosure_mcp.server.TdnetClient") as mock_class,
        ):
            mock_instance = Mock()
            mock_class.return_value = mock_instance

            client = await server_module._get_client()
            assert client is mock_instance

    @pytest.mark.asyncio
    async def test_get_client_returns_existing(self) -> None:
        existing = Mock()
        with patch.object(server_module, "_client", existing):
            client = await server_module._get_client()
            assert client is existing


class TestServerInputValidation:
    """Test server input validation patterns."""

    def test_valid_stock_code_pattern(self) -> None:
        import re

        pattern = r"^\d{4}$"
        assert re.match(pattern, "7203")
        assert re.match(pattern, "0001")
        assert not re.match(pattern, "72030")
        assert not re.match(pattern, "ABC")

    def test_valid_date_pattern(self) -> None:
        import re

        pattern = r"^\d{4}-\d{2}-\d{2}$"
        assert re.match(pattern, "2026-02-14")
        assert not re.match(pattern, "2026/02/14")


class TestServerResultConversion:
    """Test server result conversion."""

    def test_disclosure_list_to_dict(self) -> None:
        d = Disclosure(
            id="1",
            pubdate=datetime(2026, 2, 14, 15, 0),
            company_code="7203",
            company_name="トヨタ",
            title="決算短信",
            category=DisclosureCategory.EARNINGS,
            document_url="https://example.com/doc.pdf",
            exchange="東",
        )
        result = DisclosureList(total_count=1, disclosures=[d])

        data = result.to_dict()
        assert data["total_count"] == 1
        assert data["disclosures"][0]["company_code"] == "7203"
        assert data["disclosures"][0]["category"] == "earnings"

    def test_empty_result_to_dict(self) -> None:
        result = DisclosureList()
        data = result.to_dict()
        assert data["total_count"] == 0
        assert data["disclosures"] == []



"""tdnet-disclosure-mcp: TDNET timely disclosure MCP tool."""

from tdnet_disclosure_mcp.client import TdnetClient
from tdnet_disclosure_mcp.models import Disclosure, DisclosureCategory, DisclosureList

__all__ = [
    "Disclosure",
    "DisclosureCategory",
    "DisclosureList",
    "TdnetClient",
]

__version__ = "0.1.1"



"""CLI for tdnet-disclosure-mcp."""

from __future__ import annotations

import asyncio
import json
import sys
from datetime import date

import click
from loguru import logger

from tdnet_disclosure_mcp import __version__
from tdnet_disclosure_mcp.client import TdnetClient


@click.group()
@click.option("--verbose", "-v", is_flag=True, help="Enable verbose logging")
def cli(verbose: bool) -> None:
    """TDNET Disclosure MCP - Japanese timely disclosure tool."""
    if verbose:
        logger.remove()
        logger.add(sys.stderr, level="DEBUG")
    else:
        logger.remove()
        logger.add(sys.stderr, level="WARNING")


@cli.command()
def version() -> None:
    """Show version information."""
    click.echo(f"tdnet-disclosure-mcp {__version__}")


@cli.command()
@click.option("--limit", "-l", default=20, help="Maximum results (default: 20)")
@click.option("--json-output", "-j", is_flag=True, help="Output as JSON")
def latest(limit: int, json_output: bool) -> None:
    """Get latest disclosures."""

    async def _latest() -> None:
        async with TdnetClient() as client:
            result = await client.get_recent(limit=limit)

            if json_output:
                click.echo(json.dumps(result.to_dict(), ensure_ascii=False, indent=2))
            else:
                click.echo(f"Latest disclosures ({result.total_count} total)\n")
                for d in result.disclosures:
                    time_str = d.pubdate.strftime("%H:%M")
                    click.echo(f"  [{time_str}] {d.company_code} {d.company_name}")
                    click.echo(f"    {d.title}")
                    click.echo(f"    [{d.category.value}]")
                    click.echo()

    asyncio.run(_latest())


@cli.command()
@click.argument("keyword")
@click.option("--limit", "-l", default=20, help="Maximum results (default: 20)")
@click.option("--json-output", "-j", is_flag=True, help="Output as JSON")
def search(keyword: str, limit: int, json_output: bool) -> None:
    """Search disclosures by keyword."""

    async def _search() -> None:
        async with TdnetClient() as client:
            result = await client.search(keyword, limit=limit)

            if json_output:
                click.echo(json.dumps(result.to_dict(), ensure_ascii=False, indent=2))
            else:
                click.echo(f'Search results for: "{keyword}" ({result.total_count} found)\n')
                for d in result.disclosures:
                    time_str = d.pubdate.strftime("%Y-%m-%d %H:%M")
                    click.echo(f"  [{time_str}] {d.company_code} {d.company_name}")
                    click.echo(f"    {d.title}")
                    click.echo(f"    [{d.category.value}]")
                    click.echo()

    asyncio.run(_search())


@cli.command()
@click.argument("code")
@click.option("--limit", "-l", default=20, help="Maximum results (default: 20)")
@click.option("--json-output", "-j", is_flag=True, help="Output as JSON")
def company(code: str, limit: int, json_output: bool) -> None:
    """Get disclosures for a specific company (by stock code)."""

    async def _company() -> None:
        async with TdnetClient() as client:
            result = await client.get_by_code(code, limit=limit)

            if json_output:
                click.echo(json.dumps(result.to_dict(), ensure_ascii=False, indent=2))
            else:
                click.echo(f"Disclosures for {code} ({result.total_count} total)\n")
                for d in result.disclosures:
                    time_str = d.pubdate.strftime("%Y-%m-%d %H:%M")
                    click.echo(f"  [{time_str}] {d.company_name}")
                    click.echo(f"    {d.title}")
                    click.echo(f"    [{d.category.value}]")
                    click.echo()

    asyncio.run(_company())


@cli.command()
@click.argument("target_date")
@click.option("--json-output", "-j", is_flag=True, help="Output as JSON")
def by_date(target_date: str, json_output: bool) -> None:
    """Get disclosures for a specific date (YYYY-MM-DD)."""

    async def _by_date() -> None:
        try:
            parsed = date.fromisoformat(target_date)
        except ValueError:
            click.echo(
                f"Error: Invalid date format: {target_date!r}. Expected YYYY-MM-DD.",
                err=True,
            )
            sys.exit(1)

        async with TdnetClient() as client:
            result = await client.get_by_date(parsed)

            if json_output:
                click.echo(json.dumps(result.to_dict(), ensure_ascii=False, indent=2))
            else:
                click.echo(f"Disclosures for {target_date} ({result.total_count} total)\n")
                for d in result.disclosures:
                    time_str = d.pubdate.strftime("%H:%M")
                    click.echo(f"  [{time_str}] {d.company_code} {d.company_name}")
                    click.echo(f"    {d.title}")
                    click.echo()

    asyncio.run(_by_date())


@cli.command()
def test() -> None:
    """Test TDNET API connection."""

    async def _test() -> None:
        async with TdnetClient() as client:
            click.echo("Testing TDNET (Yanoshin API) connection...")

            if await client.test_connection():
                click.echo(click.style("Connection successful!", fg="green"))
            else:
                click.echo(click.style("Connection failed", fg="red"))
                sys.exit(1)

    asyncio.run(_test())


@cli.command()
@click.option("--transport", "-t", default="stdio", help="Transport type (stdio or sse)")
@click.option("--port", "-p", default=8000, help="Port for SSE transport")
def serve(transport: str, port: int) -> None:
    """Start the MCP server."""
    from tdnet_disclosure_mcp.server import mcp

    if transport == "sse":
        mcp.run(transport="sse", port=port)
    else:
        mcp.run()


if __name__ == "__main__":
    cli()



"""Domain models for tdnet-disclosure-mcp."""

from __future__ import annotations

import re
from datetime import datetime
from enum import Enum
from typing import Any

from pydantic import BaseModel, ConfigDict, Field


class DisclosureCategory(str, Enum):
    """Disclosure category types."""

    EARNINGS = "earnings"
    DIVIDEND = "dividend"
    FORECAST_REVISION = "forecast_revision"
    BUYBACK = "buyback"
    OFFERING = "offering"
    GOVERNANCE = "governance"
    OTHER = "other"


# Keywords for categorizing disclosures by title
_CATEGORY_PATTERNS: list[tuple[re.Pattern[str], DisclosureCategory]] = [
    (re.compile(r"決算短信|四半期報告|決算補足"), DisclosureCategory.EARNINGS),
    (re.compile(r"配当"), DisclosureCategory.DIVIDEND),
    (re.compile(r"業績予想.*修正|通期.*修正|予想.*変更"), DisclosureCategory.FORECAST_REVISION),
    (re.compile(r"自己株式|自社株"), DisclosureCategory.BUYBACK),
    (re.compile(r"新株|増資|公募"), DisclosureCategory.OFFERING),
    (re.compile(r"ガバナンス|役員|取締役"), DisclosureCategory.GOVERNANCE),
]


def _categorize(title: str) -> DisclosureCategory:
    """Categorize a disclosure by its title."""
    for pattern, category in _CATEGORY_PATTERNS:
        if pattern.search(title):
            return category
    return DisclosureCategory.OTHER


class Disclosure(BaseModel):
    """A single TDNET disclosure entry.

    Attributes:
        id: Unique disclosure ID.
        pubdate: Publication datetime.
        company_code: 4-digit stock code.
        company_name: Company name.
        title: Disclosure title.
        document_url: URL to the PDF document.
        xbrl_url: URL to XBRL data (may be None).
        exchange: Listed exchange(s).
        category: Auto-detected disclosure category.
        update_history: Update history text (usually empty).
    """

    id: str = Field(..., max_length=20)
    pubdate: datetime
    company_code: str = Field(..., pattern=r"^\d{4,5}$", max_length=5)
    company_name: str = Field(..., max_length=200)
    title: str = Field(..., max_length=500)
    document_url: str | None = Field(None, max_length=500)
    xbrl_url: str | None = Field(None, max_length=500)
    exchange: str = Field(default="", max_length=20)
    category: DisclosureCategory = Field(default=DisclosureCategory.OTHER)
    update_history: str | None = Field(None, max_length=500)

    model_config = ConfigDict(frozen=True)

    @classmethod
    def from_api(cls, item: dict[str, Any]) -> Disclosure:
        """Create from Yanoshin API response item."""
        tdnet = item.get("Tdnet", item)
        code_raw = tdnet.get("company_code", "")
        # Strip trailing "0" from 5-digit code
        code = code_raw[:4] if len(code_raw) == 5 else code_raw
        title = tdnet.get("title", "")

        return cls(
            id=str(tdnet.get("id", "")),
            pubdate=datetime.fromisoformat(tdnet.get("pubdate", "2000-01-01")),
            company_code=code,
            company_name=tdnet.get("company_name", "").strip(),
            title=title,
            document_url=tdnet.get("document_url"),
            xbrl_url=tdnet.get("url_xbrl"),
            exchange=tdnet.get("markets_string", ""),
            category=_categorize(title),
            update_history=tdnet.get("update_history"),
        )


class DisclosureList(BaseModel):
    """List of disclosures with metadata.

    Attributes:
        total_count: Total number of matching disclosures.
        disclosures: List of disclosure entries.
        query_date: Query date string (if date-filtered).
    """

    total_count: int = Field(0, ge=0)
    disclosures: list[Disclosure] = Field(default_factory=list)
    query_date: str | None = None

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary for MCP response."""
        return {
            "total_count": self.total_count,
            "query_date": self.query_date,
            "disclosures": [
                {
                    "id": d.id,
                    "pubdate": d.pubdate.isoformat(),
                    "company_code": d.company_code,
                    "company_name": d.company_name,
                    "title": d.title,
                    "category": d.category.value,
                    "document_url": d.document_url,
                    "exchange": d.exchange,
                }
                for d in self.disclosures
            ],
        }



"""Tests for tdnet-disclosure-mcp client."""

from datetime import date

import httpx
import pytest
import respx

from tdnet_disclosure_mcp.client import TdnetClient

_SAMPLE_RESPONSE = {
    "total_count": 2,
    "items": [
        {
            "Tdnet": {
                "id": "1001",
                "pubdate": "2026-02-14 15:30:00",
                "company_code": "72030",
                "company_name": "トヨタ自動車",
                "title": "2025年3月期 決算短信〔日本基準〕（連結）",
                "document_url": "https://example.com/doc1.pdf",
                "url_xbrl": "https://example.com/doc1.zip",
                "markets_string": "東",
                "update_history": None,
            }
        },
        {
            "Tdnet": {
                "id": "1002",
                "pubdate": "2026-02-14 16:00:00",
                "company_code": "67580",
                "company_name": "ソニーグループ",
                "title": "配当予想の修正に関するお知らせ",
                "document_url": "https://example.com/doc2.pdf",
                "url_xbrl": None,
                "markets_string": "東",
                "update_history": None,
            }
        },
    ],
}


class TestTdnetClientInit:
    """Test TdnetClient initialization."""

    def test_default_init(self) -> None:
        client = TdnetClient()
        assert client._timeout == 30.0
        assert client._http is None

    def test_custom_timeout(self) -> None:
        client = TdnetClient(timeout=60.0)
        assert client._timeout == 60.0


class TestTdnetClientRecent:
    """Test get_recent."""

    @respx.mock
    async def test_get_recent_success(self) -> None:
        respx.get("https://webapi.yanoshin.jp/webapi/tdnet/list/recent.json").mock(
            return_value=httpx.Response(200, json=_SAMPLE_RESPONSE)
        )

        async with TdnetClient() as client:
            result = await client.get_recent(limit=50)

        assert result.total_count == 2
        assert len(result.disclosures) == 2
        assert result.disclosures[0].company_code == "7203"
        assert result.disclosures[0].company_name == "トヨタ自動車"

    @respx.mock
    async def test_get_recent_empty(self) -> None:
        respx.get("https://webapi.yanoshin.jp/webapi/tdnet/list/recent.json").mock(
            return_value=httpx.Response(200, json={"total_count": 0, "items": []})
        )

        async with TdnetClient() as client:
            result = await client.get_recent()

        assert result.total_count == 0
        assert result.disclosures == []

    @respx.mock
    async def test_get_recent_limit_clamped(self) -> None:
        route = respx.get("https://webapi.yanoshin.jp/webapi/tdnet/list/recent.json").mock(
            return_value=httpx.Response(200, json={"total_count": 0, "items": []})
        )

        async with TdnetClient() as client:
            await client.get_recent(limit=999)

        assert route.called
        assert "300" in str(route.calls[0].request.url)


class TestTdnetClientByDate:
    """Test get_by_date."""

    @respx.mock
    async def test_get_by_date_success(self) -> None:
        respx.get("https://webapi.yanoshin.jp/webapi/tdnet/list/20260214.json").mock(
            return_value=httpx.Response(200, json=_SAMPLE_RESPONSE)
        )

        async with TdnetClient() as client:
            result = await client.get_by_date(date(2026, 2, 14))

        assert result.total_count == 2
        assert result.query_date == "2026-02-14"


class TestTdnetClientByDateRange:
    """Test get_by_date_range."""

    @respx.mock
    async def test_get_by_date_range_success(self) -> None:
        respx.get("https://webapi.yanoshin.jp/webapi/tdnet/list/20260210-20260214.json").mock(
            return_value=httpx.Response(200, json=_SAMPLE_RESPONSE)
        )

        async with TdnetClient() as client:
            result = await client.get_by_date_range(date(2026, 2, 10), date(2026, 2, 14))

        assert result.total_count == 2
        assert result.query_date == "2026-02-10 to 2026-02-14"


class TestTdnetClientByCode:
    """Test get_by_code."""

    @respx.mock
    async def test_get_by_code_success(self) -> None:
        respx.get("https://webapi.yanoshin.jp/webapi/tdnet/list/7203.json").mock(
            return_value=httpx.Response(200, json=_SAMPLE_RESPONSE)
        )

        async with TdnetClient() as client:
            result = await client.get_by_code("7203")

        assert result.total_count == 2
        assert result.disclosures[0].company_code == "7203"

    async def test_get_by_code_invalid(self) -> None:
        async with TdnetClient() as client:
            with pytest.raises(ValueError, match="4 digits"):
                await client.get_by_code("ABC")

    async def test_get_by_code_too_long(self) -> None:
        async with TdnetClient() as client:
            with pytest.raises(ValueError, match="4 digits"):
                await client.get_by_code("72030")


class TestTdnetClientSearch:
    """Test search."""

    @respx.mock
    async def test_search_by_name(self) -> None:
        respx.get("https://webapi.yanoshin.jp/webapi/tdnet/list/recent.json").mock(
            return_value=httpx.Response(200, json=_SAMPLE_RESPONSE)
        )

        async with TdnetClient() as client:
            result = await client.search("トヨタ")

        assert result.total_count == 1
        assert result.disclosures[0].company_code == "7203"

    @respx.mock
    async def test_search_by_code(self) -> None:
        respx.get("https://webapi.yanoshin.jp/webapi/tdnet/list/recent.json").mock(
            return_value=httpx.Response(200, json=_SAMPLE_RESPONSE)
        )

        async with TdnetClient() as client:
            result = await client.search("7203")

        assert result.total_count == 1

    @respx.mock
    async def test_search_by_title(self) -> None:
        respx.get("https://webapi.yanoshin.jp/webapi/tdnet/list/recent.json").mock(
            return_value=httpx.Response(200, json=_SAMPLE_RESPONSE)
        )

        async with TdnetClient() as client:
            result = await client.search("決算短信")

        assert result.total_count == 1
        assert "決算短信" in result.disclosures[0].title

    @respx.mock
    async def test_search_no_match(self) -> None:
        respx.get("https://webapi.yanoshin.jp/webapi/tdnet/list/recent.json").mock(
            return_value=httpx.Response(200, json=_SAMPLE_RESPONSE)
        )

        async with TdnetClient() as client:
            result = await client.search("存在しない企業")

        assert result.total_count == 0


class TestTdnetClientTestConnection:
    """Test test_connection."""

    @respx.mock
    async def test_success(self) -> None:
        respx.get("https://webapi.yanoshin.jp/webapi/tdnet/list/recent.json").mock(
            return_value=httpx.Response(200, json={"items": [], "total_count": 0})
        )

        async with TdnetClient() as client:
            assert await client.test_connection() is True

    @respx.mock
    async def test_failure(self) -> None:
        respx.get("https://webapi.yanoshin.jp/webapi/tdnet/list/recent.json").mock(
            return_value=httpx.Response(500)
        )

        async with TdnetClient() as client:
            assert await client.test_connection() is False


class TestTdnetClientLifecycle:
    """Test client lifecycle."""

    async def test_close(self) -> None:
        client = TdnetClient()
        client._get_http_client()
        assert client._http is not None

        await client.close()
        assert client._http is None

    async def test_multiple_closes(self) -> None:
        client = TdnetClient()
        await client.close()
        await client.close()

    async def test_context_manager(self) -> None:
        async with TdnetClient() as client:
            assert client is not None
        assert client._http is None


class TestTdnetClientCategoryParsing:
    """Test that API responses get correct categories."""

    @respx.mock
    async def test_categories_from_api(self) -> None:
        respx.get("https://webapi.yanoshin.jp/webapi/tdnet/list/recent.json").mock(
            return_value=httpx.Response(200, json=_SAMPLE_RESPONSE)
        )

        async with TdnetClient() as client:
            result = await client.get_recent()

        assert result.disclosures[0].category.value == "earnings"
        assert result.disclosures[1].category.value == "dividend"



"""MCP server exposing TDNET disclosure tools to LLMs via FastMCP.

Usage with Claude Desktop (add to ``claude_desktop_config.json``)::

    {
      "mcpServers": {
        "tdnet": {
          "command": "uvx",
          "args": ["tdnet-disclosure-mcp", "serve"]
        }
      }
    }

Data source: Yanoshin Web API (https://webapi.yanoshin.jp/tdnet/)
No authentication required.
"""

from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager
from datetime import date
from typing import TYPE_CHECKING, Annotated, Any

if TYPE_CHECKING:
    from collections.abc import AsyncIterator

from fastmcp import FastMCP
from pydantic import Field

from tdnet_disclosure_mcp.client import TdnetClient

# Lazily initialized client
_client: TdnetClient | None = None
_client_lock = asyncio.Lock()


async def _get_client() -> TdnetClient:
    """Return the shared TdnetClient, creating it on first call."""
    global _client
    if _client is not None:
        return _client
    async with _client_lock:
        if _client is None:
            _client = TdnetClient()
    return _client


@asynccontextmanager
async def _lifespan(server: FastMCP[dict[str, Any]]) -> AsyncIterator[dict[str, Any]]:
    """Manage TdnetClient lifecycle."""
    global _client
    yield {}
    if _client is not None:
        await _client.close()
        _client = None


mcp = FastMCP(
    name="TDNET",
    lifespan=_lifespan,
    instructions=(
        "TDNET MCP server provides tools for accessing Japanese timely disclosures "
        "(適時開示情報) from the Tokyo Stock Exchange.\n\n"
        "Available tools:\n"
        "- get_latest_disclosures: Get today's or recent disclosures\n"
        "- search_disclosures: Search by keyword or company name\n"
        "- get_company_disclosures: Get disclosures for a specific company\n"
        "- get_disclosures_by_date: Get disclosures for a specific date\n\n"
        "Disclosure categories: earnings (決算短信), dividend (配当), "
        "forecast_revision (業績予想修正), buyback (自社株買い), offering (増資), "
        "governance (ガバナンス), other.\n\n"
        "Data source: TDNET via Yanoshin Web API. No authentication required.\n"
        "Note: Only the last ~30 days of data are available."
    ),
)


@mcp.tool()
async def get_latest_disclosures(
    limit: Annotated[
        int,
        Field(description="Maximum results (default: 50, max: 300)", ge=1, le=300),
    ] = 50,
) -> dict[str, Any]:
    """Get the most recent TDNET disclosures.

    Returns the latest timely disclosures from the Tokyo Stock Exchange,
    including earnings reports (決算短信), dividend changes, forecast revisions,
    stock buybacks, and other corporate announcements.

    Each disclosure includes company code, name, title, category, and document URL.
    """
    client = await _get_client()
    result = await client.get_recent(limit=limit)
    return result.to_dict()


@mcp.tool()
async def search_disclosures(
    keyword: Annotated[
        str,
        Field(
            description="Search keyword (company name, code, or disclosure title)",
            max_length=100,
        ),
    ],
    limit: Annotated[
        int,
        Field(description="Maximum results (default: 20)", ge=1, le=100),
    ] = 20,
) -> dict[str, Any]:
    """Search TDNET disclosures by keyword.

    Searches recent disclosures by company name, stock code, or title keyword.
    Examples: "トヨタ", "7203", "決算短信", "配当"
    """
    client = await _get_client()
    result = await client.search(keyword, limit=limit)
    return result.to_dict()


@mcp.tool()
async def get_company_disclosures(
    code: Annotated[
        str,
        Field(
            description="4-digit stock code (e.g., '7203' for Toyota)",
            pattern=r"^\d{4}$",
            max_length=4,
        ),
    ],
    limit: Annotated[
        int,
        Field(description="Maximum results (default: 50)", ge=1, le=300),
    ] = 50,
) -> dict[str, Any]:
    """Get disclosures for a specific company.

    Returns all recent disclosures for the specified company,
    including earnings, dividends, forecast revisions, etc.

    Example: get_company_disclosures("7203") → Toyota's disclosures
    """
    client = await _get_client()
    result = await client.get_by_code(code, limit=limit)
    return result.to_dict()


@mcp.tool()
async def get_disclosures_by_date(
    target_date: Annotated[
        str,
        Field(
            description="Date in YYYY-MM-DD format",
            pattern=r"^\d{4}-\d{2}-\d{2}$",
            max_length=10,
        ),
    ],
) -> dict[str, Any]:
    """Get all disclosures for a specific date.

    Returns all TDNET disclosures filed on the specified date.
    Note: Only the last ~30 days are available.

    Example: get_disclosures_by_date("2026-02-14")
    """
    client = await _get_client()
    parsed_date = date.fromisoformat(target_date)
    result = await client.get_by_date(parsed_date)
    return result.to_dict()



"""Tests for tdnet-disclosure-mcp CLI."""

from datetime import datetime
from unittest.mock import AsyncMock, Mock, patch

from click.testing import CliRunner

from tdnet_disclosure_mcp import __version__
from tdnet_disclosure_mcp.cli import cli
from tdnet_disclosure_mcp.models import (
    Disclosure,
    DisclosureCategory,
    DisclosureList,
)


def _mock_result() -> DisclosureList:
    """Create a mock result for testing."""
    return DisclosureList(
        total_count=1,
        disclosures=[
            Disclosure(
                id="1",
                pubdate=datetime(2026, 2, 14, 15, 30),
                company_code="7203",
                company_name="トヨタ自動車",
                title="決算短信",
                category=DisclosureCategory.EARNINGS,
                document_url="https://example.com/doc.pdf",
                exchange="東",
            )
        ],
    )


class TestCliVersion:
    """Test version command."""

    def test_version(self) -> None:
        runner = CliRunner()
        result = runner.invoke(cli, ["version"])
        assert result.exit_code == 0
        assert __version__ in result.output


class TestCliLatest:
    """Test latest command."""

    def test_latest_success(self) -> None:
        runner = CliRunner()

        with patch("tdnet_disclosure_mcp.cli.TdnetClient") as mock_class:
            mock_client = Mock()
            mock_client.get_recent = AsyncMock(return_value=_mock_result())
            mock_client.close = AsyncMock()
            mock_client.__aenter__ = AsyncMock(return_value=mock_client)
            mock_client.__aexit__ = AsyncMock(return_value=None)
            mock_class.return_value = mock_client

            result = runner.invoke(cli, ["latest"])

        assert result.exit_code == 0
        assert "7203" in result.output
        assert "トヨタ" in result.output

    def test_latest_json(self) -> None:
        runner = CliRunner()

        with patch("tdnet_disclosure_mcp.cli.TdnetClient") as mock_class:
            mock_client = Mock()
            mock_client.get_recent = AsyncMock(return_value=_mock_result())
            mock_client.close = AsyncMock()
            mock_client.__aenter__ = AsyncMock(return_value=mock_client)
            mock_client.__aexit__ = AsyncMock(return_value=None)
            mock_class.return_value = mock_client

            result = runner.invoke(cli, ["latest", "--json-output"])

        assert result.exit_code == 0
        assert '"total_count"' in result.output


class TestCliSearch:
    """Test search command."""

    def test_search_success(self) -> None:
        runner = CliRunner()

        with patch("tdnet_disclosure_mcp.cli.TdnetClient") as mock_class:
            mock_client = Mock()
            mock_client.search = AsyncMock(return_value=_mock_result())
            mock_client.close = AsyncMock()
            mock_client.__aenter__ = AsyncMock(return_value=mock_client)
            mock_client.__aexit__ = AsyncMock(return_value=None)
            mock_class.return_value = mock_client

            result = runner.invoke(cli, ["search", "トヨタ"])

        assert result.exit_code == 0
        assert "トヨタ" in result.output


class TestCliCompany:
    """Test company command."""

    def test_company_success(self) -> None:
        runner = CliRunner()

        with patch("tdnet_disclosure_mcp.cli.TdnetClient") as mock_class:
            mock_client = Mock()
            mock_client.get_by_code = AsyncMock(return_value=_mock_result())
            mock_client.close = AsyncMock()
            mock_client.__aenter__ = AsyncMock(return_value=mock_client)
            mock_client.__aexit__ = AsyncMock(return_value=None)
            mock_class.return_value = mock_client

            result = runner.invoke(cli, ["company", "7203"])

        assert result.exit_code == 0
        assert "7203" in result.output


class TestCliByDate:
    """Test by-date command."""

    def test_by_date_success(self) -> None:
        runner = CliRunner()

        with patch("tdnet_disclosure_mcp.cli.TdnetClient") as mock_class:
            mock_result = DisclosureList(
                total_count=1,
                disclosures=[
                    Disclosure(
                        id="1",
                        pubdate=datetime(2026, 2, 14, 15, 30),
                        company_code="7203",
                        company_name="トヨタ自動車",
                        title="決算短信",
                        category=DisclosureCategory.EARNINGS,
                    )
                ],
                query_date="2026-02-14",
            )
            mock_client = Mock()
            mock_client.get_by_date = AsyncMock(return_value=mock_result)
            mock_client.close = AsyncMock()
            mock_client.__aenter__ = AsyncMock(return_value=mock_client)
            mock_client.__aexit__ = AsyncMock(return_value=None)
            mock_class.return_value = mock_client

            result = runner.invoke(cli, ["by-date", "2026-02-14"])

        assert result.exit_code == 0
        assert "2026-02-14" in result.output


class TestCliTest:
    """Test test command."""

    def test_test_success(self) -> None:
        runner = CliRunner()

        with patch("tdnet_disclosure_mcp.cli.TdnetClient") as mock_class:
            mock_client = Mock()
            mock_client.test_connection = AsyncMock(return_value=True)
            mock_client.close = AsyncMock()
            mock_client.__aenter__ = AsyncMock(return_value=mock_client)
            mock_client.__aexit__ = AsyncMock(return_value=None)
            mock_class.return_value = mock_client

            result = runner.invoke(cli, ["test"])

        assert result.exit_code == 0
        assert "successful" in result.output

    def test_test_failure(self) -> None:
        runner = CliRunner()

        with patch("tdnet_disclosure_mcp.cli.TdnetClient") as mock_class:
            mock_client = Mock()
            mock_client.test_connection = AsyncMock(return_value=False)
            mock_client.close = AsyncMock()
            mock_client.__aenter__ = AsyncMock(return_value=mock_client)
            mock_client.__aexit__ = AsyncMock(return_value=None)
            mock_class.return_value = mock_client

            result = runner.invoke(cli, ["test"])

        assert result.exit_code == 1
        assert "failed" in result.output



# Data source: TDNET via Yanoshin Web API (https://webapi.yanoshin.jp/tdnet/)
# Attribution: 適時開示情報閲覧サービス (Yanoshin)
"""Async client for TDNET disclosures via Yanoshin API.

Data source: https://webapi.yanoshin.jp/tdnet/
This uses the free Yanoshin Web API which mirrors TDNET data.
No authentication required.
"""

from __future__ import annotations

import asyncio
import re
import time as _time
from typing import TYPE_CHECKING, Any, cast

if TYPE_CHECKING:
    from datetime import date

import httpx
from loguru import logger
from pydantic import ValidationError

from tdnet_disclosure_mcp.models import Disclosure, DisclosureList

# Yanoshin API base URL
_BASE_URL = "https://webapi.yanoshin.jp/webapi/tdnet/list"

# HTTP timeout
_DEFAULT_TIMEOUT = 30.0

# Retryable HTTP status codes
_RETRYABLE_STATUS = {429, 500, 502, 503, 504}

# Max retry attempts
_MAX_RETRIES = 3

# Valid code pattern (4-digit stock code)
_VALID_CODE_RE = re.compile(r"^\d{4}$")

# Max results per request
_MAX_LIMIT = 300


class _RateLimiter:
    """Simple rate limiter using monotonic clock."""

    def __init__(self, rate: float = 1.0) -> None:
        self._interval = 1.0 / rate if rate > 0 else 0.0
        self._last: float = 0.0
        self._lock = asyncio.Lock()

    async def wait(self) -> None:
        async with self._lock:
            now = _time.monotonic()
            elapsed = now - self._last
            if elapsed < self._interval:
                await asyncio.sleep(self._interval - elapsed)
            self._last = _time.monotonic()


class TdnetClient:
    """Async client for TDNET disclosures.

    Uses the Yanoshin Web API to fetch timely disclosure data
    from TDNET (Tokyo Stock Exchange).

    No authentication required.
    """

    def __init__(self, timeout: float = _DEFAULT_TIMEOUT) -> None:
        self._timeout = timeout
        self._http: httpx.AsyncClient | None = None
        self._limiter = _RateLimiter(1.0)

    def _get_http_client(self) -> httpx.AsyncClient:
        """Get or create the HTTP client."""
        if self._http is None:
            self._http = httpx.AsyncClient(
                timeout=httpx.Timeout(self._timeout),
                headers={"Accept": "application/json"},
            )
        return self._http

    async def close(self) -> None:
        """Close the underlying HTTP client."""
        if self._http is not None:
            await self._http.aclose()
            self._http = None

    async def __aenter__(self) -> TdnetClient:
        return self

    async def __aexit__(self, *args: object) -> None:
        await self.close()

    async def _api_get(self, path: str, params: dict[str, str] | None = None) -> dict[str, Any]:
        """Make API GET request with retry and rate limiting."""
        client = self._get_http_client()
        url = f"{_BASE_URL}/{path}"
        last_exc: BaseException | None = None

        for attempt in range(_MAX_RETRIES):
            await self._limiter.wait()
            logger.debug(f"GET {url} (attempt {attempt + 1})")
            try:
                resp = await client.get(url, params=params)
                if resp.status_code in _RETRYABLE_STATUS:
                    last_exc = httpx.HTTPStatusError(
                        f"HTTP {resp.status_code}",
                        request=resp.request,
                        response=resp,
                    )
                else:
                    resp.raise_for_status()
                    return cast("dict[str, Any]", resp.json())
            except httpx.TimeoutException as e:
                last_exc = e

            if attempt < _MAX_RETRIES - 1:
                delay = 2**attempt
                logger.warning(f"Retry {attempt + 1}/{_MAX_RETRIES} for {url} after {delay}s")
                await asyncio.sleep(delay)

        raise last_exc  # type: ignore[misc]

    async def get_recent(self, limit: int = 50) -> DisclosureList:
        """Get most recent disclosures.

        Args:
            limit: Maximum number of results (1-300).

        Returns:
            List of recent disclosures.
        """
        limit = min(max(1, limit), _MAX_LIMIT)
        data = await self._api_get("recent.json", {"limit": str(limit)})
        return self._parse_response(data)

    async def get_by_date(self, target_date: date) -> DisclosureList:
        """Get disclosures for a specific date.

        Args:
            target_date: The date to query.

        Returns:
            List of disclosures for the date.
        """
        date_str = target_date.strftime("%Y%m%d")
        data = await self._api_get(f"{date_str}.json", {"limit": str(_MAX_LIMIT)})
        result = self._parse_response(data)
        result.query_date = target_date.isoformat()
        return result

    async def get_by_date_range(self, start_date: date, end_date: date) -> DisclosureList:
        """Get disclosures for a date range.

        Args:
            start_date: Start date (inclusive).
            end_date: End date (inclusive).

        Returns:
            List of disclosures in the date range.
        """
        start_str = start_date.strftime("%Y%m%d")
        end_str = end_date.strftime("%Y%m%d")
        data = await self._api_get(f"{start_str}-{end_str}.json", {"limit": str(_MAX_LIMIT)})
        result = self._parse_response(data)
        result.query_date = f"{start_date.isoformat()} to {end_date.isoformat()}"
        return result

    async def get_by_code(self, code: str, limit: int = 50) -> DisclosureList:
        """Get disclosures for a specific company.

        Args:
            code: 4-digit stock code.
            limit: Maximum number of results (1-300).

        Returns:
            List of disclosures for the company.
        """
        if not _VALID_CODE_RE.match(code):
            raise ValueError(f"Invalid stock code: {code!r} (must be 4 digits)")

        limit = min(max(1, limit), _MAX_LIMIT)
        data = await self._api_get(f"{code}.json", {"limit": str(limit)})
        return self._parse_response(data)

    async def search(self, keyword: str, limit: int = 50) -> DisclosureList:
        """Search disclosures by keyword in title.

        Fetches recent disclosures and filters by keyword.

        Args:
            keyword: Search keyword.
            limit: Maximum number of results.

        Returns:
            Filtered list of disclosures.
        """
        # Fetch recent data then filter client-side
        data = await self._api_get("recent.json", {"limit": str(_MAX_LIMIT)})
        all_disclosures = self._parse_response(data)

        keyword_lower = keyword.lower()
        filtered = [
            d
            for d in all_disclosures.disclosures
            if keyword_lower in d.title.lower()
            or keyword_lower in d.company_name.lower()
            or keyword == d.company_code
        ][:limit]

        return DisclosureList(
            total_count=len(filtered),
            disclosures=filtered,
        )

    async def test_connection(self) -> bool:
        """Test API connection.

        Returns:
            True if connection successful.
        """
        try:
            data = await self._api_get("recent.json", {"limit": "1"})
            return "items" in data
        except (httpx.HTTPError, KeyError):
            return False

    def _parse_response(self, data: dict[str, Any]) -> DisclosureList:
        """Parse API response into DisclosureList."""
        items = data.get("items", [])
        total = data.get("total_count", len(items))

        disclosures: list[Disclosure] = []
        for item in items:
            try:
                disclosures.append(Disclosure.from_api(item))
            except (ValidationError, ValueError):
                item_id = item.get("Tdnet", {}).get("id")
                logger.debug(f"Skipping invalid disclosure item: {item_id}")
                continue

        return DisclosureList(
            total_count=total,
            disclosures=disclosures,
        )



"""Tests for tdnet-disclosure-mcp models."""

from datetime import datetime

import pytest
from pydantic import ValidationError

from tdnet_disclosure_mcp.models import (
    Disclosure,
    DisclosureCategory,
    DisclosureList,
    _categorize,
)


class TestDisclosureCategory:
    """Test disclosure category detection."""

    def test_earnings(self) -> None:
        assert _categorize("2024年3月期 決算短信〔日本基準〕") == DisclosureCategory.EARNINGS

    def test_dividend(self) -> None:
        assert _categorize("配当予想の修正に関するお知らせ") == DisclosureCategory.DIVIDEND

    def test_forecast_revision(self) -> None:
        result = _categorize("業績予想の修正に関するお知らせ")
        assert result == DisclosureCategory.FORECAST_REVISION

    def test_buyback(self) -> None:
        assert _categorize("自己株式の取得に係る事項の決定") == DisclosureCategory.BUYBACK

    def test_offering(self) -> None:
        assert _categorize("新株予約権の発行に関するお知らせ") == DisclosureCategory.OFFERING

    def test_governance(self) -> None:
        assert _categorize("代表取締役の異動に関するお知らせ") == DisclosureCategory.GOVERNANCE

    def test_other(self) -> None:
        assert _categorize("定款の一部変更に関するお知らせ") == DisclosureCategory.OTHER

    def test_quarterly_report(self) -> None:
        assert _categorize("四半期報告書") == DisclosureCategory.EARNINGS

    def test_self_stock(self) -> None:
        assert _categorize("自社株買いに関するお知らせ") == DisclosureCategory.BUYBACK


class TestDisclosure:
    """Test Disclosure model."""

    def test_create_valid(self) -> None:
        d = Disclosure(
            id="12345",
            pubdate=datetime(2026, 2, 14, 15, 30),
            company_code="7203",
            company_name="トヨタ自動車",
            title="決算短信",
            category=DisclosureCategory.EARNINGS,
        )

        assert d.company_code == "7203"
        assert d.company_name == "トヨタ自動車"
        assert d.category == DisclosureCategory.EARNINGS

    def test_invalid_code(self) -> None:
        with pytest.raises(ValidationError):
            Disclosure(
                id="12345",
                pubdate=datetime(2026, 2, 14),
                company_code="ABC",
                company_name="Test",
                title="Test",
            )

    def test_frozen(self) -> None:
        d = Disclosure(
            id="12345",
            pubdate=datetime(2026, 2, 14),
            company_code="7203",
            company_name="Test",
            title="Test",
        )
        with pytest.raises(ValidationError):
            d.title = "Changed"

    def test_from_api(self) -> None:
        item = {
            "Tdnet": {
                "id": "12345",
                "pubdate": "2026-02-14 15:30:00",
                "company_code": "72030",
                "company_name": "トヨタ自動車  ",
                "title": "2025年3月期 決算短信〔日本基準〕",
                "document_url": "https://example.com/doc.pdf",
                "url_xbrl": None,
                "markets_string": "東",
                "update_history": None,
            }
        }

        d = Disclosure.from_api(item)

        assert d.id == "12345"
        assert d.company_code == "7203"
        assert d.company_name == "トヨタ自動車"
        assert d.category == DisclosureCategory.EARNINGS
        assert d.exchange == "東"

    def test_from_api_4digit_code(self) -> None:
        item = {
            "Tdnet": {
                "id": "99",
                "pubdate": "2026-01-01 09:00:00",
                "company_code": "6758",
                "company_name": "ソニー",
                "title": "お知らせ",
            }
        }

        d = Disclosure.from_api(item)
        assert d.company_code == "6758"


class TestDisclosureList:
    """Test DisclosureList model."""

    def test_create_empty(self) -> None:
        result = DisclosureList()

        assert result.total_count == 0
        assert result.disclosures == []

    def test_to_dict(self) -> None:
        d = Disclosure(
            id="1",
            pubdate=datetime(2026, 2, 14, 15, 0),
            company_code="7203",
            company_name="トヨタ",
            title="決算短信",
            category=DisclosureCategory.EARNINGS,
            document_url="https://example.com/doc.pdf",
            exchange="東",
        )
        result = DisclosureList(
            total_count=1,
            disclosures=[d],
            query_date="2026-02-14",
        )

        data = result.to_dict()

        assert data["total_count"] == 1
        assert data["query_date"] == "2026-02-14"
        assert len(data["disclosures"]) == 1
        assert data["disclosures"][0]["company_code"] == "7203"
        assert data["disclosures"][0]["category"] == "earnings"

    def test_to_dict_empty(self) -> None:
        result = DisclosureList()
        data = result.to_dict()

        assert data["total_count"] == 0
        assert data["disclosures"] == []
